from .basic_utils import *
from .misc_utils import FlattenList
from pathlib import PurePosixPath as PurePosixPath
from pathlib import Path as Path
import os
import shutil
Import("send2trash.send2trash@send2trash",globals())
Import("jsonlines",globals())

def PathToString(path="./", as_folder:bool=False):
    """Convert any representation of a path (either a `pathlib.Path` or a str) to a str, with trailing slashes for folder.

    Args:
        path: The path to be converted to string format.
        as_folder (bool): If True, force to treat it as a folder, otherwise ignored. This is useful when, for example, the desired folder does not exist.
    Returns:
        str: The result string.
    """
    path = Path(path); r = str(PurePosixPath(path/'@'))[:-1] if as_folder or path.is_dir() else str(PurePosixPath(path)); return r if r!="" else "./"

def p2s(p="./", f:bool=False):
    """Convert any representation of a path (either a `pathlib.Path` or a str) to a str, with trailing slashes for folder.

    This is a short alias for `PathToString`.

    Args:
        p: The path to be converted to string format.
        f (bool): If True, force to treat it as a folder, otherwise ignored. This is useful when, for example, the desired folder does not exist.
    Returns:
        str: The result string.
    """
    path = Path(p); r = str(PurePosixPath(path/'@'))[:-1] if f or path.is_dir() else str(PurePosixPath(path)); return r if r!="" else "./"

def p2abs(p="./"):
    """Convert a path to absolute format.

    Args:
        p: The path to be converted.
    Returns:
        str: The result string.
    """
    path = Path(p); return p2s(path.absolute())

def p2par(p="./", l:int=1):
    """Get the parent (ancestor) of a path. Note that for a relative path string, it can not trace beyond the root path.

    Args:
        p: The path to be queried.
        l (int): The number of levels of parents to jump.
    Returns:
        str: The result string.
    """
    path = Path(p)
    for _ in range(l):
        path = path.parent
    return p2s(path)

def pjoin(*plist):
    """Concatenate a list of string or paths to a path.

    Args:
        *plist (Iterable): The list to be converted.
    Returns:
        str: The result string.
    """
    path = Path()
    for p in plist:
        path = path/Path(p)
    return p2s(path)

def Prefix(path):
    """Get the prefix of a path, return the folder itself for a folder.

    Args:
        path: The path.
    Returns:
        str: The result string.
    """
    parts = p2s(path).split('.'); return '.'.join(parts[:-1] if len(parts)>1 else parts)

def Suffix(path):
    """Get the suffix of a path, return empty string "" for a folder.

    Args:
        path: The path.
    Returns:
        str: The result string.
    """
    parts = p2s(path).split('.'); return parts[-1] if len(parts)>1 else ""

def Format(path):
    """Get the format of a path.

    This is an alias for `Suffix`.

    Args:
        path: The path.
    Returns:
        str: The result string.
    """
    parts = p2s(path).split('.'); return parts[-1] if len(parts)>1 else ""

def AsFormat(path, format):
    """Get the file with the same name as path but with the desired format.

    Args:
        path: The path.
    Returns:
        str: The result string.
    """
    return Prefix(path)+'.'+format if format!="" else Prefix(path)
    
def ExistPath(path):
    """Check whether a file or folder exists.

    Args:
        path: The path.
    Returns:
        bool: True if the file or folder exists otherwise False.
    """
    return Path(path).exists()
    
def ExistFolder(path):
    """Check whether a folder exists.

    Args:
        path: The path.
    Returns:
        bool: True if the folder exists otherwise False.
    """
    return Path(path).is_dir()

def ExistFile(path):
    """Check whether a file exists.

    Args:
        path: The path.
    Returns:
        bool: True if the file exists otherwise False.
    """
    return Path(path).is_file()

def BY_NAME_CRITERIA(s):
    return s[1]
def BY_CTIME_CRITERIA(s):
    return os.path.getctime(s[1])
def BY_MTIME_CRITERIA(s):
    return os.path.getmtime(s[1])
def BY_ATIME_CRITERIA(s):
    return os.path.getatime(s[1])
def BY_SIZE_CRITERIA(s):
    return os.path.getsize(s[1])
def BUILTIN_LISTPATHS_SORT_CRITERIA():
    return {
        'BY_NAME_CRITERIA':BY_NAME_CRITERIA,
        'BY_CTIME_CRITERIA':BY_CTIME_CRITERIA,
        'BY_MTIME_CRITERIA':BY_MTIME_CRITERIA,
        'BY_ATIME_CRITERIA':BY_ATIME_CRITERIA,
        'BY_SIZE_CRITERIA':BY_SIZE_CRITERIA,
    }
def EXISTS_FILTER(s):
    return ExistPath(s[1])
def IS_FOLDER_FILTER(s):
    return ExistFolder(s[1])
def IS_FILE_FILTER(s):
    return ExistFile(s[1])
def BUILTIN_LISTPATHS_FILTER_FUNCTIONS():
    return {
        'EXISTS_FILTER':EXISTS_FILTER,
        'IS_FOLDER_FILTER':IS_FOLDER_FILTER,
        'IS_FILE_FILTER':IS_FILE_FILTER,
    }
def ListPaths(path="./", ordered:bool=False, with_path:bool=False, filter_function=None, **sort_args):
    """Return a list of subpaths under the given path.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        filter_function: Filter subpaths in format of (path, full_path). Please refer to function `BUILTIN_LISTPATHS_FILTER_FUNCTIONS()` for built-in criteria.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Please refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    path = p2s(path); subpaths = [(p2s(subpath),pjoin(path,subpath)) for subpath in os.listdir(path)]
    subpaths = list(filter(filter_function, subpaths)) if filter_function is not None else subpaths
    subpaths = sorted(subpaths, **sort_args) if ordered else subpaths
    return [subpath[with_path] for subpath in subpaths]
    
def ListFolders(path="./", ordered:bool=False, with_path=False, **sort_args):
    """Return a list of folders under the given path.

    This is an alias for `ListPaths(filter_function=IS_FOLDER_FILTER)`.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Please refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    return ListPaths(path, ordered=ordered, with_path=with_path, filter_function=IS_FOLDER_FILTER, sort_args=sort_args)
    
def ListFiles(path="./", ordered:bool=False, with_path=False, **sort_args):
    """Return a list of files under the given path.

    This is an alias for `ListPaths(filter_function=IS_FILE_FILTER)`.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Please refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    return ListPaths(path, ordered=ordered, with_path=with_path, filter_function=IS_FILE_FILTER, sort_args=sort_args)

def EnumPaths(path="./", ordered:bool=False, with_path:bool=False, filter_function=None, **sort_args):
    """Return a list of subpaths recursively under the given path.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        filter_function: Filter subpaths in format of (path, full_path). Notice that the filter function will only be applied at the base level instead of applied recursively. Please refer to function `BUILTIN_LISTPATHS_FILTER_FUNCTIONS()` for built-in criteria.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Notice that the sort function will only be applied at the base level instead of applied recursively. PPlease refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    path = p2s(path); files = ListFiles(path); folders = ListFolders(path)
    subfiles = [[pjoin(folder,f) for f in EnumPaths(pjoin(path,folder))] for folder in folders]
    subpaths = [(p2s(subpath),pjoin(path,subpath)) for subpath in files+folders+FlattenList(subfiles)]
    subpaths = list(filter(filter_function, subpaths)) if filter_function is not None else subpaths
    subpaths = sorted(subpaths, **sort_args) if ordered else subpaths
    return [subpath[with_path] for subpath in subpaths]

def EnumFolders(path="./", ordered:bool=False, with_path:bool=False, **sort_args):
    """Return a list of folders recursively under the given path.

    This is an alias for `ListPaths(filter_function=IS_FOLDER_FILTER)`.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        filter_function: Filter subpaths in format of (path, full_path). Notice that the filter function will only be applied at the base level instead of applied recursively. Please refer to function `BUILTIN_LISTPATHS_FILTER_FUNCTIONS()` for built-in criteria.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Notice that the sort function will only be applied at the base level instead of applied recursively. PPlease refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    return EnumPaths(path, ordered=ordered, with_path=with_path, filter_function=IS_FOLDER_FILTER, sort_args=sort_args)

def EnumFiles(path="./", ordered:bool=False, with_path:bool=False, **sort_args):
    """Return a list of files recursively under the given path.

    This is an alias for `ListPaths(filter_function=IS_FILE_FILTER)`.

    Args:
        path: The path.
        ordered (bool): If True, sort subpaths by criteria specified in `sort_args`, otherwise in original order from `os.listdir()`.
        with_path (bool): If True, pre-attach the given path to all subpaths, otherwise ignored.
        filter_function: Filter subpaths in format of (path, full_path). Notice that the filter function will only be applied at the base level instead of applied recursively. Please refer to function `BUILTIN_LISTPATHS_FILTER_FUNCTIONS()` for built-in criteria.
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Notice that the sort function will only be applied at the base level instead of applied recursively. PPlease refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.
    Returns:
        List[str]: The result strings.
    """
    return EnumPaths(path, ordered=ordered, with_path=with_path, filter_function=IS_FILE_FILTER, sort_args=sort_args)
    
def CreateFolder(path):
    """Create the given path as folder, parents will be automatically built.

    Args:
        path: The path.
    Returns:
        bool: True if the folder is created, otherwise already exists.
    """
    if not ExistPath(path):
        Path(path).mkdir(parents=True,exist_ok=False); return True
    else:
        return False
    
def CreateFile(path):
    """Create the given path as file, parents will be automatically built.

    Args:
        path: The path.
    Returns:
        bool: True if the file is created, otherwise already exists.
    """
    if not ExistPath(path):
        Path(p2par(path)).mkdir(parents=True,exist_ok=True)
        if Suffix(path)!='':
            Path(path).touch(exist_ok=False)
        return True
    else:
        return False

def Delete(path, rm=False):
    """Delete the given path.

    Args:
        path: The path.
        rm (bool): If True, use `shutil` to enforce remove, otherwise `send2trash` only.
    Returns:
        bool: True if the path is deleted, otherwise it does not exist.
    """
    if ExistPath(path):
        path = p2s(path); (shutil.rmtree(path) if ExistFolder(path) else os.remove(path)) if rm else send2trash(path); return True
    else:
        return False

def ClearFolder(path, rm=False):
    """Clear the given folder if it exists, create if it does not exist.

    Args:
        path: The path.
        rm (bool): If True, use `shutil` to enforce remove, otherwise `send2trash` only.
    Returns:
        None
    """
    Delete(path,rm=rm); CreateFolder(path)

def ClearFile(path, rm=False):
    """Clear the given file if it exists, create if it does not exist.

    Args:
        path: The path.
        rm (bool): If True, use `shutil` to enforce remove, otherwise `send2trash` only.
    Returns:
        None
    """
    Delete(path,rm=rm); CreateFile(path)

def CopyFile(src, dst, rm=False):
    """Copy file from src to dst, notice that dst will be deleted if exists.

    Args:
        src: The source path.
        dst: The destination path.
        rm (bool): If True, use `shutil` to enforce remove (if `dst` exists), otherwise `send2trash` only.
    Returns:
        None
    """
    Delete(dst,rm=rm); CreateFile(dst); shutil.copyfile(p2s(src), p2s(dst))

def ReplaceFile(src, dst, rm=False):
    """Replace file from src to dst, notice that dst will be deleted if exists.

    This is an alias for `CopyFile`.

    Args:
        src: The source path.
        dst: The destination path.
        rm (bool): If True, use `shutil` to enforce remove (if `dst` exists), otherwise `send2trash` only.
    Returns:
        None
    """
    Delete(dst,rm=rm); CreateFile(dst); shutil.copyfile(p2s(src), p2s(dst))

def CopyFolder(src, dst, rm=False):
    """Copy folder from src (folder name included) to dst (folder name included). Only existing files will be deleted if exists.

    This is NOT the same as `ReplaceFolder`, which does not merge `src` to `dst`, instead, it deletes the entire `dst` directory if exists.

    Args:
        src: The source path.
        dst: The destination path.
        rm (bool): If True, use `shutil` to enforce remove (if conflict arises), otherwise `send2trash` only.
    Returns:
        None
    """
    for file in EnumFiles(src):
        CopyFile(pjoin(src,file),pjoin(dst,file),rm=rm)

def ReplaceFolder(src, dst, rm=False):
    """Copy folder from src (folder name included) to dst (folder name included). The entire dst directory will be deleted if exists.

    This is NOT the same as `CopyFolder`, which merges `src` to `dst`.

    Args:
        src: The source path.
        dst: The destination path.
        rm (bool): If True, use `shutil` to enforce remove (if `dst` exists), otherwise `send2trash` only.
    Returns:
        None
    """
    CreateFolder(dst); Delete(dst, rm=rm); shutil.copytree(p2s(src,f=True), p2s(dst,f=True))



from .file_utils import *
import io
Import("torch",globals())

def SaveTorch(obj, path, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True):
    """Save an object as torch file.

    Args:
        obj: The object to be saved.
        path: The save path.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        None
    """
    CreateFile(path); path = p2s(path)
    torch.save(obj, path, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)

def LoadTorch(path, map_location=None):
    """Load an object from existing torch file.

    Args:
        path: The load path.
        map_location: The `map_location` argument for `torch.load`.
    Returns:
        Any: The loaded object.
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    return torch.load(path, map_location=map_location)

def DumpsTorch(obj, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True)
    """Save an object as torch bytearray.

    Args:
        obj: The object to be saved.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        None
    """
    buff = io.BytesIO()
    torch.save(obj, buff, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)
    buff.seek(0); return buff.read()


import sys
from importlib import import_module
from subprocess import Popen, PIPE, STDOUT
from typing import Union, Literal, Optional, List, Dict

def CMD(command:str, wait:bool=True, shell:bool=True, sudo:bool=False):
    """Call system cmd console to execute a command (`subprocess.Popen`).

    Args:
        command (str): the command to be executed. Should be a str of can be converted to str.
        wait (bool): wait until the command finishes and return the handle or directly return the handle.
        shell (bool): whether to display shell for `subprocess.Popen`.
        sudo (bool): whether to pre-attach "sudo" in front of the command.
    Returns:
        bool: Return the handle created by `subprocess.Popen`.
    """
    sudo_command = ("sudo " if sudo else "")+str(command)
    h = Popen(sudo_command,shell=shell)
    if wait:
        h.wait()
    return h

def BUILTIN_PIP_SOURCES():
    return {
        "": "",
        "aliyun": "http://mirrors.aliyun.com/pypi/simple/",
        "douban": "http://pypi.douban.com/simple/",
        "tuna": "http://pypi.tuna.tsinghua.edu.cn/simple/",
        "ustc": "http://pypi.mirrors.ustc.edu.cn/simple/",
    }
def PIP(package:str, source:str="", pip3:bool=True, upgrade:bool=False, force:bool=False, force_deps:bool=False, https:bool=False, args:str=""):
    """Install a python package by pip (or pip3), with built-in sources and other settings wrapped. Please refer to function `BUILTIN_PIP_SOURCES()` for built-in sources.

    Args:
        package (str): The package name to be installed.
        source (str): "--index-url" and "--trusted-host" setting. There are some built-in sources, but you could also specify your own sources by passing the url. The url should be valid and starts with "http://" or "https://".
        pip3 (bool): If True, use "pip3", otherwise "pip".
        upgrade (bool): If True, use "--upgrade", otherwise ignored.
        force (bool): If True, use "--force-reinstall", otherwise ignored.
        force_deps (bool): If False and "force" is set to True, use "--no-deps", otherwise ignored.
        https (bool): If True, replace built-in sources to corresponding https version, it does NOT work for custom sources.
        args (str): Custom args for pip (or pip3) to be appended.
    Returns:
        None
    """
    package = str(package).lower()  # Pypi is case insensitive
    index_url = BUILTIN_PIP_SOURCES()[source].replace("http://", "https://" if https else "http://") if source in BUILTIN_PIP_SOURCES() else source
    trusted_host = index_url.split("https://" if https else "http://")[-1].split("/")[0]
    source_command = f"--index-url {index_url} --trusted-host {trusted_host} " if source!="" else ""
    upgrade_command = "--upgrade " if upgrade else ""
    reinstall_command = "--force-reinstall " if force else ""
    nodeps_command = "--no-deps " if (force and not force_deps) else ""
    pip_command = "pip3" if pip3 else "pip"
    CMD(f"{pip_command} install {package} {source_command}{upgrade_command}{reinstall_command}{nodeps_command}{args}",wait=True)

def __import_unsafe__(name:str, alias:str, module_globals:Dict, force_reimport:bool=False):
    if force_reimport or (alias not in module_globals):
        if (alias is None) and ("." in name):
            __import_unsafe__(name.split('.')[0], name.split('.')[0], module_globals=module_globals, force_reimport=force_reimport)
        else:
            if alias is None:
                alias = name
            try:
                module = import_module(name)
                module_globals.update({alias:module})
                return
            except:
                pass
            
            try:
                if "." in name:
                    parent, mod = ".".join(name.split('.')[:-1]), name.split('.')[-1]
                    module = import_module(parent).__dict__[mod]
                    module_globals.update({alias:module})
                    return
            except:
                pass
            
            raise ImportError(name)

def USEFUL_PYTHON_PACKAGES_BUILTIN():
    """List of useful packages in python standard libraries, encoded as a comma separated string."""
    return "re,os,sys,pdb,glob,math,time,json,copy,types,random,shutil,string,pickle,zipfile,marshal,logging,pathlib,"\
        +  "argparse,itertools,subprocess,collections,PIL.Image.Image@Image,collections.defaultdict@defaultdict,"\
        +  "collections.abc.Iterable@iterable,copy.copy@copy,copy.deepcopy@deepcopy,pathlib.Path@Path,"\
        +  "pathlib.PurePosixPath@PurePosixPath"
def USEFUL_PYTHON_PACKAGES_EXTRA():
    """List of useful packages NOT in python standard libraries, encoded as a comma separated string."""
    return "tqdm,jsonlines,requests,sklearn,numpy@np,pandas@pd,seaborn@sns,matplotlib@mpl,matplotlib.pyplot@plt"
def USEFUL_PYTHON_PACKAGES_TORCH():
    """List of useful packages in pytorch, encoded as a comma separated string."""
    return "torch,torch,torch.nn@nn,torch.nn.init@nninit,torch.nn.functional@F,torch.optim@optim,torch.optim.lr_scheduler@sched,torch.utils.data@TUD,torch.utils.data.DataLoader@DataLoader,torch.utils.data.Dataset@Dataset,torch.nn.parallel.DistributedDataParallel@DDP"

def Import(packages:Union[List[str],str], module_globals, mode:Literal['strict','install_strict','install_ignore','ignore']='ignore', force_reimport:bool=False, **install_args):
    """Dynamically import packages.

    Args:
        packages (str or List[str]): The package or packages to be imported. Use "@" for alias. Order matters: please place parent packages before their children.
        module_globals: The `globals()` of the module that performs the import. You should almost always put `module_globals=globals()` here.
        mode (str): If set to "strict", failures directly result in `ImportError`.
                    If set to "install_strict", try to install missing packages using `PIP` with arguments in `install_args`, and then failures result in `ImportError`.
                    If set to "install_ignore", try to install missing packages using `PIP` with arguments in `install_args`, and then failures are ignored.
                    If set to "ignore", failures are ignored.
        force_reimport (bool): If True, force import even if the module is already imported, otherwise ignored.
        install_args: Arguments for calling `PIP`. Only useful for "install_strict" or "install_ignore" mode.
    Returns:
        List[str]: Packages or functions that failed to be imported.
    """
    packages = [package for package in packages] if not isinstance(packages, str) else packages.split(','); failures = []; errors = []
    for package in packages:
        if package!="":
            failed = False; error = None
            if package.count('@')>1:
                failed = True
            else:
                name, alias = package.split('@') if package.count('@')==1 else (package, None)
                try:
                    __import_unsafe__(name, alias, module_globals=module_globals, force_reimport=force_reimport)
                    failed = False
                except Exception as e:
                    if "install" in mode:
                        try:
                            PIP(name, **install_args)
                            __import_unsafe__(name, alias, module_globals=module_globals, force_reimport=force_reimport)
                            failed = False
                        except:
                            failed = True; error = e
                    else:
                        failed = True; error = e
            if failed:
                failures.append(package)
                errors.append(error)
    if ("strict" in mode) and len(failures)>0:
        raise ImportError(list(zip(failures, errors)))
    else:
        return failures

def ForName(cls:str):
    """Return a class by class name.

    Args:
        cls (str): Class name.
    Returns:
        type: The class.
    """
    return sys.modules[cls]


from .file_utils import *
import json
import pickle
Import("demjson",globals())
Import("jsonlines",globals())
Import("simplejson",globals())
Import("jsonpickle",globals())

def BUILTIN_JSON_BACKENDS():
    return ['json','jsonl','demjson','simplejson','jsonpickle']
def BUILTIN_JSON_PRETTIFY_BACKENDS():
    return ['json',        'demjson','simplejson','jsonpickle']
def SaveJson(obj, path, backend:Literal['json','jsonl','demjson','simplejson','jsonpickle']='json', indent:Optional[int]=None, append:bool=False, *args, **kwargs):
    """Save an object as json (or jsonl) file.

    Args:
        obj: The object to be saved.
        path: The save path.
        backend (str): Specify backend for saving an object in json format. Please refer to function `BUILTIN_JSON_BACKENDS()` for built-in backends.
        indent (int/None): The `indent` argument for saving in json format, only works if backend is not "jsonl".
        append (bool): If True, use "a" mode instead of "w" mode, only works if backend is "jsonl".
    Returns:
        None
    """
    CreateFile(path); path = p2s(path)
    if backend=='jsonl':
        assert (indent is None), ("'jsonl' format does not support parameter 'indent'!")
        with jsonlines.open(path, "a" if append else "w") as f:
            for data in obj:
                f.write(data)
    else:
        assert (append is False), ("'json' format does not support parameter 'append'!")
        module = globals()[backend]
        with open(path, "w") as f:
            if backend in ['json','simplejson']:
                module.dump(obj, f, indent=indent, *args, **kwargs)
            else:
                f.write(module.dumps(obj, indent=indent, *args, **kwargs))

def LoadJson(path, backend:Literal['json','jsonl','demjson','simplejson','picklejson']='json'):
    """Load an object from existing json (or jsonl) file.

    Args:
        path: The load path.
        backend (str): Specify backend for loading an object in json format. Please refer to function `BUILTIN_JSON_BACKENDS()` for built-in backends.
    Returns:
        Any: The loaded object.
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    if backend=='jsonl':
        with open(path, "r") as f:
            return [data for data in jsonlines.Reader(f)]
    else:
        module = globals()[backend]
        with open(path, "r") as f:
            if backend in ['json','simplejson']:
                module.load(f)
            else:
                module.loads(f.read())

def PrettifyJson(path, load_backend:Literal['json','demjson','simplejson','picklejson']='json', save_backend:Literal['json','demjson','simplejson','picklejson']='json', indent:Optional[int]=4, *args, **kwargs):
    """Load and re-save a existing json file.

    Args:
        path: The json file path.
        load_backend (str): Specify backend for loading an object in json format. Please refer to function `BUILTIN_JSON_PRETTIFY_BACKENDS()` for built-in backends.
        save_backend (str): Specify backend for saving an object in json format. Please refer to function `BUILTIN_JSON_PRETTIFY_BACKENDS()` for built-in backends.
        indent (int/None): The `indent` argument for saving in json format.
    Returns:
        None
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    SaveJson(LoadJson(path, backend=load_backend), path, backend=save_backend, indent=indent, *args, **kwargs)

def SavePickle(obj, path, protocol:Optional[int]=None):
    """Save an object as pickle file.

    Args:
        obj: The object to be saved.
        path: The save path.
        protocol (int/None): The `protocol` argument for `pickle.dump`.
    Returns:
        None
    """
    CreateFile(path); path = p2s(path)
    with open(path, "wb") as f:
        pickle.dump(obj, f, protocol=protocol)

def LoadPickle(path):
    """Load an object from existing pickle file.

    Args:
        path: The load path.
    Returns:
        Any: The loaded object.
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    return pickle.load(open(PathToString(path), "rb"))

def DumpsPickle(obj, protocol:Optional[int]=None):
    """Save an object as pickle bytearray.

    Args:
        obj: The object to be saved.
        protocol (int/None): The `protocol` argument for `pickle.dump`.
    Returns:
        None
    """
    return pickle.dumps(obj, protocol=protocol)